//
//  character.cpp
//  Lily
//
//  Created by KMY on 2015/05/06.
//  Copyright (c) 2015年 KMY. All rights reserved.
//

#include "character.h"

bool Character::isHalf(lchar c) {
	return !Character::isFull_utf8(c);
}

bool Character::isFull(lchar c) {
	return isFull_utf8(c);
}

int Character::fullCount() {
	return fullCount_utf8();
}

bool Character::isAscii(lchar c) {
	return c <= 0x7f;
}

bool Character::isFull_utf8(lchar c) {
	return !isAscii(c);
}

int Character::fullCount_utf8() {
	return 3;
}


//
// lchar <-> char 変換
//

int Character::toLchar(const char *from, lchar *to) {
	
	// 半角文字
	if (isHalf(from[0])) {
		*to = from[0];
		return 1;
	}
	
	// 全角文字
	else {
		*to = 0;
		for (int i=0; i<fullCount(); i++) {
			*to <<= sizeof(char) * 8;
			*to |= static_cast<unsigned char>(from[i]);
		}
		return fullCount();
	}
}

void Character::toLchar(const char *from, lchar *to, int lcharLen, int charLen) {
	
	// 文字列の長さ
	if (!charLen) charLen = strlen(from);
	
	// コピー
	int fi=0, ti=0;
	while (fi < charLen &&
		   ti < lcharLen-1 &&						// NULL用に１つ空けておく
		   from[fi]									// 原文がNULLでないか
		   ) {
		lchar c;
		int b = toLchar(&from[fi], &c);
		to[ti++] = c;
		fi += b;
	}
	
	// NULL文字
	to[ti] = 0;
}

void Character::toLcharn(const char *from, lchar **to, int charLen) {
	
	// 文字列の長さ
	if (!charLen) charLen = strlen(from);
	
	// コピー先の領域を確保
	int llen = lstrlen(from);
	*to = new lchar[llen+1];
	
	// コピー
	toLchar(from, *to, llen+1);
}

int Character::toChar(lchar from, char *to) {
	
	// lcharを1バイトずつ食って、ゼロでない部分のみをコピー
	int count = 0;
	char* cp = reinterpret_cast<char*>(&from+1);
	for (int i=0; i<sizeof(lchar); i++) {
		cp--;
		if (*cp) {
			to[count++] = *cp;
		}
	}
	return count;
}

void Character::toChar(const lchar *from, char *to, int charLen, int lcharLen) {
	
	// 文字列の長さ
	if (!lcharLen) lcharLen = lstrlen(from);
	
	// コピー
	int fi=0, ti=0;
	while (fi < lcharLen &&
		   ti < charLen-1 &&				// NULL文字分の領域を確保
		   from[fi]							// NULL文字まできたらおわり
		   ) {
		ti += Character::toChar(from[fi], to+ti);
		fi++;
	}
	
	// NULL文字
	to[ti] = 0;
}

void Character::toCharn(const lchar *from, char **to, int lcharLen) {
	
	// 文字列の長さ
	if (!lcharLen) lcharLen = lstrlen(from);
	
	// コピー先の領域を確保
	int len = strlen(from);
	*to = new char[len+1];
	
	// コピー
	toChar(from, *to, len+1);
}


//
// lchar 文字列関連
//

void Character::lstrcpy(const lchar *from, lchar *to, int maxlen) {
	
	// 文字列の長さをはがる
	if (!maxlen) maxlen = lstrlen(from);
	
	// コピー
	while (maxlen--) {
		*to = *from;
		to++;
		from++;
	}
	*to = 0;				// NULL文字
}

void Character::lstrcpyn(const lchar *from, lchar **to, int maxlen) {
	
	// 文字列の長さをはがる
	if (!maxlen) maxlen = lstrlen(from);
	
	// 配列をヒープ領域に作成
	lchar *heap = new lchar[maxlen+1];
	
	// コピー
	int count = 0;
	while (maxlen--) {
		*(heap+count) = *from;
		count++;
		from++;
	}
	*(heap+count) = 0;		// NULL文字
	*to = heap;
}

int Character::lstrlen(const lchar *string) {
	int count = 0;
	while (string[count++]);
	return count-1;
}

int Character::lstrlen(const char *string) {
	int count = 0;
	int len = strlen(string);
	
	// サイズを取得（全角・半角にわけて処理）
	for (int i=0; i<len; i++) {
		if (isHalf(string[i])) count++;
		else {
			count++;
			i += fullCount() - 1;
		}
	}
	
	return count;
}


//
// char 文字列関連
//

void Character::strcpy(const char *from, char *to, int maxlen) {
	
	// 文字列の長さをはがる
	if (!maxlen) maxlen = strlen(from);
	
	// コピー
	while (maxlen--) {
		*to = *from;
		to++;
		from++;
	}
	*to = 0;			// NULL文字
}

void Character::strcpyn(const char *from, char **to, int maxlen) {
	
	// 文字列の長さをはがる
	if (!maxlen) maxlen = strlen(from);
	
	// 配列をヒープ領域に作成
	char *heap = new char[maxlen+1];
	
	// コピー
	int count = 0;
	while (maxlen--) {
		*(heap+count) = *from;
		count++;
		from++;
	}
	*(heap+count) = 0;		// NULL文字
	
	*to = heap;
}

int Character::strlen(const char *str) {
	int count = 0;
	while (str[count++]);
	return count-1;
}

int Character::strlen(const lchar *str) {
	int count = 0;
	int i = 0;
	while (lchar c = str[i++]) {
		for (int j=0; j<sizeof(lchar); j++) {
			if (*(reinterpret_cast<char*>(&c)+j)) {
				count++;
			}
		}
	}
	return count;
}


//
// 等価であるか
//

bool Character::equal(const char *str1, const char *str2) {
	int i = 0;
	while (str1[i]) {
		if (str1[i] != str2[i]) return false;
		i++;
	}
	return (str2[i]) ? false : true;
}

bool Character::equal(const char *str1, const lchar *str2) {
	char* str2_char;
	Character::toCharn(str2, &str2_char);
	bool result = Character::equal(str1, str2_char);
	delete str2_char;
	return result;
}

bool Character::equal(const lchar *str1, const char *str2) {
	return Character::equal(str2, str1);
}

bool Character::equal(const lchar *str1, const lchar *str2) {
	int i = 0;
	while (str1[i]) {
		if (str1[i] != str2[i]) return false;
		i++;
	}
	return (str2[i]) ? false : true;
}

bool Character::equal(char c1, char c2) {
	return c1 == c2;
}

bool Character::equal(const char *c1, lchar c2) {
	lchar c1_l;
	Character::toLchar(c1, &c1_l);
	return c1_l == c2;
}

bool Character::equal(lchar c1, const char *c2) {
	return Character::equal(c2, c1);
}

bool Character::equal(lchar c1, lchar c2) {
	return c1 == c2;
}
